import shlex
import click
import importlib

import keyring
import github3

from . import __url__, harvest
from .jira import JIRA, JIRAError
from .utils import cached_property, taskstatus
from .git import Repository


class NullIntegrationHelper:
    def register(self, *args, **kwargs):
        pass

    def close(self):
        pass


class WarnIntegrationHelper(NullIntegrationHelper):
    def __init__(self):
        self._shown = False

    def register(self, *args, **kwargs):
        if not self._shown:
            self._shown = True
            click.secho('')
            click.secho('  Lancet executable called directly', fg='yellow')
            click.secho('  ---------------------------------', fg='yellow')
            click.secho('')
            click.secho('  Setup the shell integration to enjoy some of the')
            click.secho('  super powers we built right into lancet.')
            click.secho('')
            click.secho('  This basically means to add the following snippet')
            click.secho('  to your shell initialization file:')
            click.secho('')
            click.secho('    lancet --setup-helper | source /dev/stdin')
            click.secho('')
            click.secho('  See {} for addtional details.'.format(
                click.style('https://lancet.rtd.org', fg='green')))
            click.secho('')


class ShellIntegrationHelper(NullIntegrationHelper):
    def __init__(self, filename):
        self.filename = filename
        self.fh = open(filename, 'w')

    def register(self, *args, raw=False):
        cmd = args[0] if raw else ' '.join(shlex.quote(a) for a in args)
        self.fh.write(cmd)
        self.fh.write('\n')

    def close(self):
        self.fh.close()


class Lancet:
    def __init__(self, config, integration_helper):
        self.config = config
        self.integration_helper = integration_helper

    def get_instance_from_config(self, section, key):
        import_path = self.config.get(section, key)
        module_path, callable_name = import_path.rsplit('.', 1)
        factory = getattr(importlib.import_module(module_path), callable_name)
        return factory(self)

    def defer_to_shell(self, *args, **kwargs):
        return self.integration_helper.register(*args, **kwargs)

    def get_credentials(self, service, checker=None):
        url = self.config.get(service, 'url')
        username = self.config.get(service, 'username')
        key = 'lancet+{}'.format(url)
        if username:
            password = keyring.get_password(key, username)
            if password:
                return url, username, password

        with taskstatus.suspend():
            while True:
                click.echo(
                    'Please provide your authentication information for {}'
                    .format(url)
                )
                if not username:
                    username = click.prompt('Username')
                else:
                    click.echo('Username: {}'.format(username))
                password = click.prompt('Password', hide_input=True)

                if checker:
                    with taskstatus('Checking provided credentials') as ts:
                        if not checker(url, username, password):
                            ts.fail('Login failed')
                            username, password = None, None
                            continue
                        else:
                            ts.ok('Correctly authenticated to {}', url)

                keyring.set_password(key, username, password)
                return url, username, password

    def get_issue(self, key=None):
        # TODO: Move this method to the JIRA class

        if key is None:
            name_getter = self.get_instance_from_config(
                'repository', 'branch_name_getter')
            key = name_getter.get_issue_key(self.repo.head.name)
        elif key.isdigit():
            project_key = self.config.get('tracker', 'default_project')
            if project_key:
                key = '{}-{}'.format(project_key, key)
        return self.tracker.issue(key)

    @cached_property
    def repo(self):
        # Can be cleared like this: self.__class__.repo.fget.cache_clear()
        return Repository('./.git')

    @cached_property
    def github(self):
        url = self.config.get('github', 'url')
        # TODO: This is only used to create the key, but we shall add support
        # github enterprise as well
        key = 'lancet+{}'.format(url)
        username = self.config.get('github', 'username')
        token = keyring.get_password(key, username)

        if not token:
            def two_factor_callback():
                with taskstatus.suspend():
                    return click.prompt('2-factor auth code')

            with taskstatus.suspend():
                while True:
                    click.echo(
                        'Please provide your authentication information for {}'
                        .format(url)
                    )
                    if not username:
                        username = click.prompt('Username')
                    else:
                        click.echo('Username: {}'.format(username))
                    password = click.prompt('Password', hide_input=True)

                    with taskstatus('Getting authorization token') as ts:
                        scopes = ['user', 'repo']
                        try:
                            auth = github3.authorize(
                                username, password, scopes, 'Lancet', __url__,
                                two_factor_callback=two_factor_callback
                            )
                        except github3.GitHubError as e:
                            ts.fail('Login failed ({})', e)
                            username, password = None, None
                            continue
                        else:
                            ts.ok('New token correctly generated')
                            break

                token = '{}:{}'.format(auth.id, auth.token)
                keyring.set_password(key, username, token)

        id, token = token.split(':', 1)
        gh = github3.login(token=token)
        self.call_on_close(gh._session.close)
        return gh

    @cached_property
    def tracker(self):
        def checker(url, username, password):
            try:
                JIRA(options={'server': url}, basic_auth=(username, password))
            except JIRAError:
                return False
            else:
                return True

        url, username, password = self.get_credentials('tracker', checker)
        tracker = JIRA(options={'server': url},
                       basic_auth=(username, password))
        self.call_on_close(tracker.close)
        return tracker

    @cached_property
    def timer(self):
        return harvest.client_factory(self)



import os
import sys
import re
import glob
import configparser
import pdb

import click
import github3
import keyring
from giturlparse import parse as giturlparse

from . import __version__
from .settings import load_config, USER_CONFIG, LOCAL_CONFIG, PROJECT_CONFIG
from .git import BranchGetter
from .base import Lancet, WarnIntegrationHelper, ShellIntegrationHelper
from .utils import taskstatus


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def hr(char='─', width=None, **kwargs):
    if width is None:
        width = click.get_terminal_size()[0]
    click.secho('─' * width, **kwargs)


def get_issue(lancet, key=None):
    with taskstatus('Looking up issue on the issue tracker') as ts:
        issue = lancet.get_issue(key)
        summary = issue.fields.summary
        crop = len(summary) > 40
        if crop:
            summary = summary[:40] + '...'
        ts.ok('Retrieved issue {}: {}'.format(issue.key, summary))
    return issue


def get_transition(ctx, lancet, issue, to_status):
    current_status = issue.fields.status.name
    if current_status != to_status:
        transitions = [t['id'] for t in lancet.tracker.transitions(issue)
                       if t['to']['name'] == to_status]
        if not transitions:
            click.secho(
                'No transition from "{}" to "{}" found, aborting.'
                .format(current_status, to_status),
                fg='red', bold=True
            )
            ctx.exit(1)
        elif len(transitions) > 1:
            click.secho(
                'Multiple transitions found from "{}" to "{}", aborting.'
                .format(current_status, to_status),
                fg='red', bold=True
            )
            ctx.exit(1)
        else:
            transition_id = transitions[0]
    else:
        transition_id = None
    return transition_id


def assign_issue(lancet, issue, username, active_status=None):
    with taskstatus('Assigning issue to you') as ts:
        assignee = issue.fields.assignee
        if not assignee or assignee.name != username:
            if issue.fields.status.name == active_status:
                ts.abort('Issue already active and not assigned to you')
            else:
                lancet.tracker.assign_issue(issue, username)
                ts.ok('Issue assigned to you')
        else:
            ts.ok('Issue already assigned to you')


def set_issue_status(lancet, issue, to_status, transition):
    with taskstatus('Setting issue status to "{}"'.format(to_status)) as ts:
        if transition is not None:
            lancet.tracker.transition_issue(issue, transition)
            ts.ok('Issue status set to "{}"'.format(to_status))
        else:
            ts.ok('Issue already "{}"'.format(to_status))


def setup_helper(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    base = os.path.abspath(os.path.dirname(__file__))
    helper = os.path.join(base, 'helper.sh')
    with open(helper) as fh:
        click.echo(fh.read())
    ctx.exit()


def get_branch(lancet, issue, base_branch=None, create=True):
    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')
    remote_name = lancet.config.get('repository', 'remote_name')

    remote = lancet.repo.lookup_remote(remote_name)
    credentials = lancet.repo.get_credentials_for_remote(remote)

    name_getter = lancet.get_instance_from_config(
        'repository', 'branch_name_getter')
    branch_getter = BranchGetter(base_branch, credentials, name_getter,
                                 remote_name)

    return branch_getter(lancet.repo, issue, create=create)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__, message='%(prog)s %(version)s')
@click.option('-d', '--debug/--no-debug', default=False)
@click.option('--setup-helper', callback=setup_helper, is_flag=True,
              expose_value=False, is_eager=True,
              help='Print the shell integration code and exit.')
@click.pass_context
def main(ctx, debug):
    # TODO: Enable this using a command line switch
    # import logging
    # logging.basicConfig(level=logging.DEBUG)

    if debug:
        def exception_handler(type, value, traceback):
            click.secho('\nAn exception occurred while executing the '
                        'requested command:', fg='red')
            hr(fg='red')
            sys.__excepthook__(type, value, traceback)

            click.secho('\nAs requested I will now drop you inside an '
                        'interactive debugging session:', fg='red')
            hr(fg='red')

            pdb.post_mortem(traceback)
        sys.excepthook = exception_handler

    try:
        integration_helper = ShellIntegrationHelper(
            os.environ['LANCET_SHELL_HELPER'])
    except KeyError:
        integration_helper = WarnIntegrationHelper()

    if os.path.exists(PROJECT_CONFIG):
        config = load_config(PROJECT_CONFIG)
    else:
        config = load_config()

    ctx.obj = Lancet(config, integration_helper)
    ctx.obj.call_on_close = ctx.call_on_close
    ctx.call_on_close(integration_helper.close)


@click.command()
@click.option('-k', '--key', 'method', flag_value='key', default=True)
@click.option('-d', '--dir', 'method', flag_value='dir')
@click.argument('project')
@click.pass_obj
def activate(lancet, method, project):
    """Switch to this project."""
    workspace = os.path.expanduser(lancet.config.get('lancet', 'workspace'))
    project_path = None

    with taskstatus('Looking up project') as ts:
        if method == 'key':
            config_files = glob.glob(
                os.path.join(workspace, '*', LOCAL_CONFIG))

            for path in config_files:
                config = load_config(path)
                key = config.get('tracker', 'default_project', fallback=None)

                if key.lower() == project.lower():
                    project_path = os.path.dirname(path)
        elif method == 'dir':
            project_path = os.path.join(workspace, project)

        if not project_path or not os.path.exists(project_path):
            ts.abort('Project "{}" not found (using {}-based lookup)',
                     project, method)

    # Load the configuration
    config = load_config(os.path.join(project_path, LOCAL_CONFIG))

    # cd to the project directory
    lancet.defer_to_shell('cd', project_path)

    # Activate virtualenv
    venv = config.get('lancet', 'virtualenv', fallback=None)
    if venv:
        venv_path = os.path.join(project_path, os.path.expanduser(venv))
        activate_script = os.path.join(venv_path, 'bin', 'activate')
        lancet.defer_to_shell('source', activate_script)
    else:
        if 'VIRTUAL_ENV' in os.environ:
            lancet.defer_to_shell('deactivate')

main.add_command(activate)


@click.command()
@click.option('--base', '-b', 'base_branch')
@click.argument('issue')
@click.pass_context
def workon(ctx, issue, base_branch):
    """
    Start work on a given issue.

    This command retrieves the issue from the issue tracker, creates and checks
    out a new aptly-named branch, puts the issue in the configured active,
    status, assigns it to you and starts a correctly linked Harvest timer.

    If a branch with the same name as the one to be created already exists, it
    is checked out instead. Variations in the branch name occuring after the
    issue ID are accounted for and the branch renamed to match the new issue
    summary.

    If the `default_project` directive is correctly configured, it is enough to
    give the issue ID (instead of the full project prefix + issue ID).
    """
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    active_status = lancet.config.get('tracker', 'active_status')
    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')

    # Get the issue
    issue = get_issue(lancet, issue)

    # Get the working branch
    branch = get_branch(lancet, issue, base_branch)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, active_status)

    # Make sure the issue is assigned to us
    assign_issue(lancet, issue, username, active_status)

    # Activate environment
    set_issue_status(lancet, issue, active_status, transition)

    with taskstatus('Checking out working branch') as ts:
        lancet.repo.checkout(branch.name)
        ts.ok('Checked out working branch based on "{}"'.format(base_branch))

    with taskstatus('Starting harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Started harvest timer')

main.add_command(workon)


@click.command()
@click.argument('issue')
@click.pass_obj
def time(lancet, issue):
    """
    Start an Harvest timer for the given issue.

    This command takes care of linking the timer with the issue tracker page
    for the given issue.
    """
    issue = get_issue(lancet, issue)

    with taskstatus('Starting harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Started harvest timer')

main.add_command(time)


@click.command()
@click.pass_context
def pause(ctx):
    """
    Pause work on the current issue.

    This command puts the issue in the configured paused status and stops the
    current Harvest timer.
    """
    lancet = ctx.obj
    paused_status = lancet.config.get('tracker', 'paused_status')

    # Get the issue
    issue = get_issue(lancet)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, paused_status)

    # Activate environment
    set_issue_status(lancet, issue, paused_status, transition)

    with taskstatus('Pausing harvest timer') as ts:
        lancet.timer.pause()
        ts.ok('Harvest timer paused')

main.add_command(pause)


@click.command()
@click.pass_context
def resume(ctx):
    """
    Resume work on the currently active issue.

    The issue is retrieved from the currently active branch name.
    """
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    active_status = lancet.config.get('tracker', 'active_status')

    # Get the issue
    issue = get_issue(lancet)

    # Make sure the issue is in a correct status
    transition = get_transition(ctx, lancet, issue, active_status)

    # Make sure the issue is assigned to us
    assign_issue(lancet, issue, username, active_status)

    # Activate environment
    set_issue_status(lancet, issue, active_status, transition)

    with taskstatus('Resuming harvest timer') as ts:
        lancet.timer.start(issue)
        ts.ok('Resumed harvest timer')

main.add_command(resume)


@click.command(name='pr')
@click.option('--base', '-b', 'base_branch')
@click.option('-o', '--open-pr/--no-open-pr', default=False)
@click.pass_context
def pull_request(ctx, base_branch, open_pr):
    """Create a new pull request for this issue."""
    lancet = ctx.obj

    username = lancet.config.get('tracker', 'username')
    review_status = lancet.config.get('tracker', 'review_status')
    remote_name = lancet.config.get('repository', 'remote_name')

    if not base_branch:
        base_branch = lancet.config.get('repository', 'base_branch')

    # Get the issue
    issue = get_issue(lancet)

    transition = get_transition(ctx, lancet, issue, review_status)

    # Get the working branch
    branch = get_branch(lancet, issue, create=False)

    with taskstatus('Checking pre-requisites') as ts:
        if not branch:
            ts.abort('No working branch found')

        assignee = issue.fields.assignee
        if not assignee or assignee.name != username:
            ts.abort('Issue currently not assigned to you')

        # TODO: Check mergeability

    # TODO: Check remote status (PR does not already exist)

    # Push to remote
    with taskstatus('Pushing to "{}"', remote_name) as ts:
        remote = lancet.repo.lookup_remote(remote_name)
        if not remote:
            ts.abort('Remote "{}" not found', remote_name)

        remote.credentials = lancet.repo.get_credentials_for_remote(remote)
        remote.push(branch.name)

        ts.ok('Pushed latest changes to "{}"', remote_name)

    # Create pull request
    with taskstatus('Creating pull request') as ts:
        p = giturlparse(remote.url)
        gh_repo = lancet.github.repository(p.owner, p.repo)

        message = click.edit("{} – {}\n\n{}".format(
            issue.key, issue.fields.summary, issue.permalink()))

        if not message:
            ts.abort('You didn\'t provide a title for the pull request')

        title, body = message.split('\n', 1)
        title = title.strip()

        if not title:
            ts.abort('You didn\'t provide a title for the pull request')

        try:
            pr = gh_repo.create_pull(title, base_branch, branch.branch_name,
                                     body.strip('\n'))
        except github3.GitHubError as e:
            if len(e.errors) == 1:
                error = e.errors[0]
                if 'pull request already exists' in error['message']:
                    ts.ok('Pull request does already exist')
                else:
                    ts.abort('Could not create pull request ({})',
                             error['message'])
            else:
                raise
        else:
            ts.ok('Pull request created at {}', pr.html_url)

    # Update issue
    set_issue_status(lancet, issue, review_status, transition)

    # TODO: Post to activity stream on JIRA
    # TODO: Post to HipChat?

    # Stop harvest timer
    with taskstatus('Pausing harvest timer') as ts:
        lancet.timer.pause()
        ts.ok('Harvest timer paused')

    # Open the pull request page in the browser if requested
    if open_pr:
        click.launch(pr.html_url)

main.add_command(pull_request)


@click.command()
@click.argument('issue', required=False)
@click.pass_obj
def browse(lancet, issue):
    """
    Open the issue tracker page for the given issue in your default browser.

    If no issue is provided, the one linked to the current branch is assumed.
    """
    click.launch(get_issue(lancet, issue).permalink())

main.add_command(browse)


@click.command()
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def setup(ctx, force):
    """Wizard to create the user-level configuration file."""
    if os.path.exists(USER_CONFIG) and not force:
        click.secho(
            'An existing configuration file was found at "{}".\n'
            .format(USER_CONFIG),
            fg='red', bold=True
        )
        click.secho(
            'Please remove it before in order to run the setup wizard or use\n'
            'the --force flag to overwrite it.'
        )
        ctx.exit(1)

    click.echo('Address of the issue tracker (your JIRA instance). \n'
               'Normally in the form https://<company>.atlassian.net.')
    tracker_url = click.prompt('URL')
    tracker_user = click.prompt('Username for {}'.format(tracker_url))
    click.echo()

    click.echo('Address of the time tracker (your Harvest instance). \n'
               'Normally in the form https://<company>.harvestapp.com.')
    timer_url = click.prompt('URL')
    timer_user = click.prompt('Username for {}'.format(timer_url))
    click.echo()

    config = configparser.ConfigParser()

    config.add_section('tracker')
    config.set('tracker', 'url', tracker_url)
    config.set('tracker', 'username', tracker_user)

    config.add_section('harvest')
    config.set('harvest', 'url', timer_url)
    config.set('harvest', 'username', timer_user)

    with open(USER_CONFIG, 'w') as fh:
        config.write(fh)

    click.secho('Configuration correctly written to "{}".'
                .format(USER_CONFIG), fg='green')

    # TODO: Add wizard to setup shell integration

main.add_command(setup)


@click.command()
@click.option('-f', '--force/--no-force', default=False)
@click.pass_context
def init(ctx, force):
    """Wizard to create a project-level configuration file."""
    if os.path.exists(PROJECT_CONFIG) and not force:
        click.secho(
            'An existing configuration file was found at "{}".\n'
            .format(PROJECT_CONFIG),
            fg='red', bold=True
        )
        click.secho(
            'Please remove it before in order to run the setup wizard or use\n'
            'the --force flag to overwrite it.'
        )
        ctx.exit(1)

    project_key = click.prompt('Project key on the issue tracker')
    base_branch = click.prompt('Integration branch', default='master')

    virtualenvs = ('.venv', '.env', 'venv', 'env')
    for p in virtualenvs:
        if os.path.exists(os.path.join(p, 'bin', 'activate')):
            venv = p
            break
    else:
        venv = ''
    venv_path = click.prompt('Path to virtual environment', default=venv)

    project_id = click.prompt('Project ID on Harvest', type=int)
    task_id = click.prompt('Task id on Harvest', type=int)

    config = configparser.ConfigParser()

    config.add_section('lancet')
    config.set('lancet', 'virtualenv', venv_path)

    config.add_section('tracker')
    config.set('tracker', 'default_project', project_key)

    config.add_section('harvest')
    config.set('harvest', 'project_id', str(project_id))
    config.set('harvest', 'task_id', str(task_id))

    config.add_section('repository')
    config.set('repository', 'base_branch', base_branch)

    with open(PROJECT_CONFIG, 'w') as fh:
        config.write(fh)

    click.secho('\nConfiguration correctly written to "{}".'
                .format(PROJECT_CONFIG), fg='green')

main.add_command(init)


@click.command()
@click.argument('service', required=False)
@click.pass_obj
def logout(lancet, service):
    """Forget saved passwords for the web services."""
    if service:
        services = [service]
    else:
        services = ['tracker', 'harvest']

    for service in services:
        url = lancet.config.get(service, 'url')
        key = 'lancet+{}'.format(url)
        username = lancet.config.get(service, 'username')
        with taskstatus('Logging out from {}', url) as ts:
            if keyring.get_password(key, username):
                keyring.delete_password(key, username)
                ts.ok('Logged out from {}', url)
            else:
                ts.ok('Already logged out from {}', url)

main.add_command(logout)


@click.command(name='harvest-projects')
@click.argument('query', required=False)
@click.pass_obj
def harvest_projects(lancet, query):
    """List Harvest projects, optionally filtered with a regexp."""
    projects = lancet.timer.projects()

    if query:
        regexp = re.compile(query, flags=re.IGNORECASE)

        def match(project):
            match = regexp.search(project['name'])
            if match is None:
                return False
            project['match'] = match
            return True
        projects = (p for p in projects if match(p))

    for project in sorted(projects, key=lambda p: p['name'].lower()):
        name = project['name']

        if 'match' in project:
            m = project['match']
            s, e = m.start(), m.end()
            match = click.style(name[s:e], fg='green')
            name = name[:s] + match + name[e:]

        click.echo('{:>9d} {} {}'.format(
            project['id'], click.style('‣', fg='yellow'), name))

main.add_command(harvest_projects)


@click.command(name='harvest-tasks')
@click.argument('project_id', type=int)
@click.pass_obj
def harvest_tasks(lancet, project_id):
    """List Harvest tasks for the given project ID."""
    projects = lancet.timer.projects()

    for project in projects:
        if project['id'] == project_id:
            click.echo('{:>9d} {} {}'.format(
                project['id'], click.style('‣', fg='yellow'), project['name']))
            hr()

            for task in project['tasks']:
                click.echo('{:>9d} {} {}'.format(
                    task['id'], click.style('‣', fg='yellow'), task['name']))
            break

main.add_command(harvest_tasks)


# TODO:
# * review
#     pull
#     ci-status
#     pep8
#     diff
#     mergeability (rebase is of the submitter responsibility)
# * merge
#     pull, merge, delete
# * issues
#     list all open/assigned issues (or by filter)
# * comment
#     adds a comment to the currently active issue



"""
From http://en.wikipedia.org/wiki/Scalpel:

    A scalpel, or lancet, is a small and extremely sharp bladed instrument used
    for surgery, anatomical dissection, and various arts and crafts (called a
    hobby knife).
"""

__author__ = 'Jonathan Stoppani'
__email__ = 'jonathan@stoppani.name'
__version__ = '0.6.0'
__url__ = 'https://github.com/GaretJax/lancet'



import functools

from html.parser import HTMLParser
from urllib.parse import urljoin, urlencode
import requests


class MetaRetriever(HTMLParser):
    def __init__(self, name, *args, **kwargs):
        self.meta_name = name
        self.meta_content = ''
        kwargs['convert_charrefs'] = True
        super().__init__(*args, **kwargs)

    def handle_starttag(self, tag, attrs):
        attrs = dict(attrs)
        if tag == 'meta' and attrs.get('name', None) == self.meta_name:
            self.meta_content = attrs.get('content', '')


def get_meta_content(html, name):
    retriever = MetaRetriever(name)
    retriever.feed(html)
    return retriever.meta_content


def requires_login(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        self._login()
        return func(self, *args, **kwargs)
    return wrapper


class HarvestError(Exception):
    pass


class HarvestAPI:
    def __init__(self, server, basic_auth):
        self.server = server
        self.auth = basic_auth
        self._projects = None
        self._session = requests.Session()
        self._session.auth = basic_auth
        self._session.headers = {
            'accept': 'application/json'
        }

    def _get(self, url):
        r = self._session.get(urljoin(self.server, url))
        payload = r.json()
        if r.status_code != 200:
            raise HarvestError(payload['message'])
        return payload

    def _post(self, url, data):
        return self._session.post(urljoin(self.server, url)).json()

    def toggle(self, id):
        return self._get('daily/timer/{}'.format(id))

    def pause(self, id=None):
        for entry in self.daily():
            if 'timer_started_at' in entry:
                if id is None or id == entry['id']:
                    self.toggle(entry['id'])
                break
        else:
            # No running timers
            pass

    def whoami(self):
        return self._get('account/who_am_i')

    def projects(self):
        if not self._projects:
            self._projects = self._get('daily')['projects']
        return self._projects

    def tasks(self, project_id):
        project_id = int(project_id)
        for project in self.projects():
            if project['id'] == int(project_id):
                return project['tasks']

    def daily(self):
        daily = self._get('daily')
        self._projects = daily['projects']
        return daily['day_entries']

    def close(self):
        self._session.close()


class HarvestPlatform(HarvestAPI):
    platform_url = 'https://platform.harvestapp.com'

    def __init__(self, server, basic_auth, project_id_getter, task_id_getter):
        self._web_session = requests.Session()
        self.get_project_id = project_id_getter
        self.get_task_id = task_id_getter
        self._csrf_token = None
        super().__init__(server, basic_auth)

    def _get_csrf_token(self):
        if not self._csrf_token:
            r = self._web_session.get(urljoin(self.server, '/account/login'))
            self._csrf_token = get_meta_content(r.text, 'csrf-token')
        return self._csrf_token

    def _login(self):
        # TODO: Cache the harvest session across commands invocations
        if '_harvest_sess' in self._web_session.cookies:
            return
        data = {
            'utf8': '✓',
            'authenticity_token': self._get_csrf_token(),
            'user[email]': self.auth[0],
            'user[password]': self.auth[1],
        }
        self._web_session.post(
            urljoin(self.server, '/account/create_session'),
            data=data,
        )
        self._csrf_token = None

    @requires_login
    def start(self, issue, resume=True):
        project = issue.fields.project

        if resume:
            for entry in reversed(self.daily()):
                ext = entry.get('external_ref', None)
                if not ext:
                    continue
                if ext['group_id'] != str(project.id):
                    continue
                if ext['id'] != str(issue.id):
                    continue
                if 'timer_started_at' not in entry:
                    self.toggle(entry['id'])
                return

        name = '{} - {}'.format(issue.key, issue.fields.summary)
        project_id = self.get_project_id(self, issue)
        task_id = self.get_task_id(self, project_id, issue)

        qs = urlencode({
            'app_name': 'JIRA',
            'base_url': issue.permalink(),
            'external_account_id': 1,  # ID of the JIRA instance
            'external_group_id': project.id,
            'external_group_name': project.name,
            'external_item_id': issue.id,
            'external_item_name': name,
            'service': 'divio-ch.atlassian.net',
        })
        data = {
            'utf8': '✓',
            'project_id': project_id,
            'task_id': task_id,
            'notes': name,
            'hours': '',
            'button': '',
        }
        headers = {
            'x-csrf-token': self._get_csrf_token(),
            'x-requested-with': 'XMLHttpRequest',
        }
        r = self._web_session.post(
            urljoin(self.platform_url, '/platform/timer?{}'.format(qs)),
            data=data,
            headers=headers,
        )
        assert r.status_code == 200
        assert 'message' in r.json()

    def close(self):
        self._web_session.close()
        super(HarvestPlatform, self).close()


class MappedProjectID:
    def __init__(self, project_ids):
        self._project_ids = project_ids

    @classmethod
    def fromstring(cls, string):
        if not string:
            return cls({})
        project_ids = string.split(',')
        project_ids = (p.split(':') for p in project_ids)
        project_ids = ((None, p[0]) if len(p) == 1 else p for p in project_ids)
        project_ids = ((k, int(v)) for k, v in project_ids if v)
        project_ids = dict(project_ids)
        return cls(project_ids)

    def __call__(self, timer, issue):
        try:
            return self._project_ids[str(issue.fields.issuetype)]
        except KeyError:
            pass

        try:
            return self._project_ids[None]
        except KeyError:
            pass

        raise ValueError('Could not find a project ID for issue type "{}"'
                         .format(issue.fields.issuetype))


def mapped_project_id_getter(lancet):
    return MappedProjectID.fromstring(
        lancet.config.get('harvest', 'project_id'))


def fixed_task_id_getter(lancet):
    def getter(timer, project_id, issue):
        return int(lancet.config.get('harvest', 'task_id'))
    return getter


class EpicTaskMapper:
    def __init__(self, epic_link_field, epic_name_field):
        self.epic_link_field = epic_link_field
        self.epic_name_field = epic_name_field

    def get_epic(self, issue):
        from jira.resources import Issue
        if issue.fields.issuetype.subtask:
            parent = Issue(issue._options, issue._session)
            parent.find(issue.fields.parent.key)
            issue = parent
        epic = Issue(issue._options, issue._session)
        epic.find(getattr(issue.fields, self.epic_link_field))
        return epic

    def __call__(self, timer, project_id, issue):
        try:
            epic = self.get_epic(issue)
        except:
            raise ValueError('Could not find the epic for task {}'.format(
                issue.key))
        epic_name = getattr(epic.fields, self.epic_name_field)

        for t in timer.tasks(project_id):
            if t['name'] == epic_name:
                return t['id']

        raise ValueError(
            'Could not find a task with the name "{}" in the Harvest project '
            'with ID {}'.format(epic_name, project_id))


def epic_task_id_getter(lancet):
    return EpicTaskMapper(lancet.config.get('harvest', 'epic_link_field'),
                          lancet.config.get('harvest', 'epic_name_field'))


def credentials_checker(url, username, password):
    """Check the provided credentials using the Harvest API."""
    api = HarvestAPI(url, (username, password))
    try:
        api.whoami()
    except HarvestError:
        return False
    else:
        return True


def client_factory(lancet):
    """Construct a new Harvest client."""
    url, username, password = lancet.get_credentials(
        'harvest', credentials_checker)

    project_id_getter = lancet.get_instance_from_config(
        'harvest', 'project_id_getter')
    task_id_getter = lancet.get_instance_from_config(
        'harvest', 'task_id_getter')

    client = HarvestPlatform(
        server=url,
        basic_auth=(username, password),
        project_id_getter=project_id_getter,
        task_id_getter=task_id_getter
    )
    lancet.call_on_close(client.close)
    return client



from jira.client import GreenHopper  # NOQA
from jira.exceptions import JIRAError

from requests.cookies import RequestsCookieJar


__all__ = ['JIRA', 'JIRAError']


class ObliviousCookieJar(RequestsCookieJar):
    def set_cookie(self, *args, **kwargs):
        """Simply ignore any request to set a cookie."""
        pass

    def copy(self):
        """Make sure to return an instance of the correct class on copying."""
        return ObliviousCookieJar()


class JIRA(GreenHopper):
    def _create_http_basic_session(self, username, password):
        super(JIRA, self)._create_http_basic_session(username, password)

        # XXX: JIRA logs the web user out if we send the session cookies we get
        # back from the first request in any subsequent requests. As we don't
        # need cookies when accessing the API anyway, just ignore all of them.
        self._session.cookies = ObliviousCookieJar()

    def close(self):
        self._session.close()



import functools
import sys
import curses
import click


def cached_property(*args, **kwargs):
    return property(functools.lru_cache()(*args, **kwargs))


class PrintTaskStatus:
    _active_tasks = []

    def __init__(self, msg, *args, **kwargs):
        self.setup()
        self._done = False
        self.msg = msg.format(*args, **kwargs)

    @classmethod
    def setup(cls):
        if getattr(cls, '_setup', False):
            return
        curses.setupterm()
        cls.BOL = curses.tigetstr('cr')
        cls.CLEAR_EOL = curses.tigetstr('el')
        cls._setup = True

    @classmethod
    def clear_line(cls):
        sys.stdout.buffer.write(cls.BOL + cls.CLEAR_EOL)

    def __enter__(self):
        PrintTaskStatus._active_tasks.append(self)
        msg = ' {}  {}...'.format(
            click.style('*', fg='yellow', blink=True), self.msg)
        click.echo(msg, nl=False)
        return self

    def __exit__(self, type, value, tb):
        PrintTaskStatus._active_tasks.pop()
        if not self._done:
            if type:
                self.fail('{}: {}', type.__name__, value)
            else:
                self.clear_line()

    def ok(self, msg, *args, **kwargs):
        self.clear_line()
        msg = msg.format(*args, **kwargs)
        click.echo(' {}  {}'.format(click.style('✓', fg='green'), msg))
        self._done = True

    def fail(self, msg, *args, **kwargs):
        self.clear_line()
        msg = msg.format(*args, **kwargs)
        click.echo(' {}  {}'.format(click.style('✗', fg='red'), msg))
        self._done = True

    def abort(self, msg, *args, **kwargs):
        msg += ', aborting'
        self.fail(msg, *args, **kwargs)
        sys.exit(1)

    @classmethod
    def suspend(cls):
        return SuspendTask(cls._active_tasks)


class SuspendTask:
    def __init__(self, stack):
        self.stack = stack

    def __enter__(self):
        if self.stack:
            task = self.stack[-1]
            if not task._done:
                task.clear_line()

    def __exit__(self, type, value, tb):
        if self.stack:
            task = self.stack.pop()
            if not task._done:
                task.__enter__()

taskstatus = PrintTaskStatus



import sys
import re
import six

import pygit2
import click
from slugify import slugify
from giturlparse import parse as giturlparse

from .utils import taskstatus


TOKEN_USER = b'x-oauth-basic'


class PrefixedIDBranchName:
    slug_length = 50

    def get_prefix(self, issue):
        raise NotImplementedError()

    def get_issue_key(self, branch_name):
        raise Exception('Unable to find current issue.')

    def __call__(self, issue):
        discriminator = '{}{}_'.format(self.get_prefix(issue), issue.key)
        slug = slugify(issue.fields.summary[:self.slug_length])
        full_name = '{}{}'.format(discriminator, slug)
        return discriminator, full_name


class FixedPrefixIDBranchName(PrefixedIDBranchName):
    def __init__(self, prefix):
        self._prefix = prefix

    def get_prefix(self, issue):
        return self._prefix

    def get_issue_key(self, branch_name):
        match = re.search(self._prefix + r'([A-Z]{2,}-[0-9]+)', branch_name)
        if match is None:
            raise Exception('Unable to find current issue.')
        return match.group(1)


class TaskTypePrefixIDBranchName(PrefixedIDBranchName):
    def __init__(self, prefixes):
        self._prefixes = prefixes

    def get_prefix(self, issue):
        try:
            return self._prefixes[str(issue.fields.issuetype)]
        except KeyError:
            pass

        try:
            return self._prefixes[None]
        except KeyError:
            pass

        raise ValueError('Could not find a prefix for issue type "{}"'
                         .format(issue.fields.issuetype))

    def get_issue_key(self, branch_name):
        for prefix in six.itervalues(self._prefixes):
            match = re.search(prefix + r'([A-Z]{2,}-[0-9]+)', branch_name)
            if match is not None:
                return match.group(1)
        else:
            raise Exception('Unable to find current issue.')

    @classmethod
    def fromstring(cls, string):
        prefixes = string.split(',')
        prefixes = (p.split(':') for p in prefixes)
        prefixes = ((None, p[0]) if len(p) == 1 else p for p in prefixes)
        prefixes = dict(prefixes)
        return cls(prefixes)


def prefixed_id_branch_name(lancet):
    prefix = lancet.config.get('repository', 'branch_name_prefix')
    return TaskTypePrefixIDBranchName.fromstring(prefix)


class BranchGetter:
    def __init__(self, base_branch, remote_credentials, branch_name_getter,
                 remote_name='origin'):
        self.base_branch = base_branch
        self.remote_name = remote_name
        self.remote_credentials = remote_credentials
        self.get_branch_name = branch_name_getter

    def get_base_branch(self, repo):
        return repo.lookup_branch(
            '{}/{}'.format(self.remote_name, self.base_branch),
            pygit2.GIT_BRANCH_REMOTE
        )

    def get_branch(self, repo, issue, from_remote=False):
        discriminator, full_name = self.get_branch_name(issue)

        if from_remote:
            branch_type = pygit2.GIT_BRANCH_REMOTE
            discriminator = '{}/{}'.format(self.remote_name, discriminator)
        else:
            branch_type = pygit2.GIT_BRANCH_LOCAL

        branches = [b for b in repo.listall_branches(branch_type)
                    if b.startswith(discriminator)]

        if len(branches) > 1:
            click.secho('Multiple matching branches found!',
                        fg='red', bold=True)
            click.echo()
            click.echo('The prefix {} matched the following branches:'
                       .format(discriminator))
            click.echo()
            for b in branches:
                click.echo(' {} {}'.format(click.style('*', fg='red'), b))
            click.echo()
            click.echo('Please remove all but one in order to continue.')
            sys.exit(1)
        elif branches:
            branch = repo.lookup_branch(branches[0], branch_type)
            if from_remote:
                branch = repo.create_branch(full_name, branch.get_object())
            elif branch.branch_name != full_name:
                # We only need to rename if we're working on an already
                # existing local branch
                branch.rename(full_name)
                branch = repo.lookup_branch(full_name)
            return branch

    def __call__(self, repo, issue, create=True):
        branch = self.get_branch(repo, issue)

        if not branch:
            # No local branches found, fetch remote
            with taskstatus('Fetching from "{}"', self.remote_name) as ts:
                remote = repo.lookup_remote(self.remote_name)
                if not remote:
                    ts.abort('Remote "{}" not found', self.remote_name)

                remote.credentials = self.remote_credentials
                remote.fetch()
                ts.ok('Fetched latest changes from "{}"', self.remote_name)

            # Check remote branches
            with taskstatus('Creating working branch') as ts:
                branch = self.get_branch(repo, issue, from_remote=True)

                # Create off origin base branch
                if branch:
                    ts.ok('Created new working branch based on existing '
                          'remote branch')
                elif create:
                    _, full_name = self.get_branch_name(issue)

                    base = self.get_base_branch(repo)
                    if not base:
                        ts.abort('Base branch "{}" not found on remote "{}"',
                                 self.base_branch, self.remote_name)
                    branch = repo.create_branch(full_name, base.get_object())
                    ts.ok('Created new working branch')

        return branch


class Repository(pygit2.Repository):
    def lookup_remote(self, name):
        for remote in self.remotes:
            if remote.name == name:
                return remote

    def get_credentials_for_remote(self, remote):
        if not remote:
            return
        p = giturlparse(remote.url)
        remote_username = p._user

        if p.protocol == 'ssh':
            credentials = pygit2.KeypairFromAgent(remote_username)
        elif p.protocol == 'https':
            # TODO: What if this fails? (platform, pwd not stored,...)
            try:
                import subprocess
                out = subprocess.check_output([
                    'security', 'find-internet-password',
                    '-r', 'htps',
                    '-s', p.domain,
                    '-g',
                ], stderr=subprocess.STDOUT)

                username = re.search(rb'"acct"<blob>="([^"]+)"', out)
                username = username.group(1)

                password = re.search(rb'password: "([^"]+)"', out)
                password = password.group(1)

                if password == TOKEN_USER:
                    username, password = password, username
            except:
                raise NotImplementedError('No authentication support.')

            credentials = pygit2.UserPass(username, password)

        return credentials



"""
Configuration management for the different components.
"""

import os
import configparser


PACKAGE = 'lancet'
LOCAL_CONFIG = '.{}'.format(PACKAGE)
DEFAULT_CONFIG = os.path.join(os.path.dirname(__file__),
                              'default-settings.ini')
SYSTEM_CONFIG = '/etc/{0}/{0}.conf'.format(PACKAGE)
USER_CONFIG = os.path.expanduser(os.path.join('~', LOCAL_CONFIG))
PROJECT_CONFIG = os.path.join(os.path.realpath('.'), LOCAL_CONFIG)

DEFAULT_FILES = [
    DEFAULT_CONFIG,
    SYSTEM_CONFIG,
    USER_CONFIG,
]


def load_config(path=None, defaults=None):
    """
    Loads and parses an INI style configuration file using Python's built-in
    configparser module. If path is specified, load it.
    If ``defaults`` (a list of strings) is given, try to load each entry as a
    file, without throwing any error if the operation fails.
    If ``defaults`` is not given, the following locations listed in the
    DEFAULT_FILES constant are tried.
    To completely disable defaults loading, pass in an empty list or ``False``.
    Returns the SafeConfigParser instance used to load and parse the files.
    """

    if defaults is None:
        defaults = DEFAULT_FILES

    config = configparser.ConfigParser(allow_no_value=True)

    if defaults:
        config.read(defaults)

    if path:
        with open(path) as fh:
            config.read_file(fh)

    return config
